@@ -28,10 +28,9 @@
 
  */
 
-/* $Id: apc_cache.c,v 3.175 2008/12/02 14:45:20 pajoye Exp $ */
+/* $Id: apc_cache.c 286717 2009-08-03 05:25:32Z gopalv $ */
 
 #include "apc_cache.h"
-#include "apc_lock.h"
 #include "apc_sma.h"
 #include "apc_main.h"
 #include "apc_globals.h"
@@ -55,24 +54,7 @@ static unsigned int hash(apc_cache_key_t key)
 /* }}} */
 
 /* {{{ string_nhash_8 */
-static unsigned int string_nhash_8(const char *s, size_t len)
-{
-    register const unsigned int *iv = (const unsigned int *)s;
-    register unsigned int h = 0;
-    register const unsigned int *e  = (const unsigned int *)(s + len - (len % sizeof(unsigned int)));
-
-    for(;iv<e;iv++) {
-        h += *iv;
-        h = (h << 7) | (h >> ((8*sizeof(unsigned int)) - 7));
-    }
-    s = (const char *)iv;
-    for(len %= sizeof(unsigned int);len;len--) {
-        h += *(s++);
-    }
-    h ^= (h >> 13);
-    h ^= (h >> 7);
-    return h;
-}
+#define string_nhash_8(s,len) (unsigned int)(zend_inline_hash_func(s, len))
 /* }}} */
 
 /* {{{ make_slot */
@@ -283,6 +265,8 @@ void apc_cache_clear(apc_cache_t* cache)
         cache->slots[i] = NULL;
     }
 
+    memset(&cache->header->lastkey, 0, sizeof(apc_keyid_t));
+
     cache->header->busy = 0;
     CACHE_UNLOCK(cache);
 }
@@ -295,7 +279,7 @@ static void apc_cache_expunge(apc_cache_t* cache, size_t size)
     time_t t;
     TSRMLS_FETCH();
 
-    t = sapi_get_request_time(TSRMLS_C);
+    t = apc_time();
 
     if(!cache) return;
 
@@ -340,7 +324,7 @@ static void apc_cache_expunge(apc_cache_t* cache, size_t size)
                  */
                 if((*p)->value->type == APC_CACHE_ENTRY_USER) {
                     if((*p)->value->data.user.ttl) {
-                        if((*p)->creation_time + (*p)->value->data.user.ttl < t) {
+                        if((time_t) ((*p)->creation_time + (*p)->value->data.user.ttl) < t) {
                             remove_slot(cache, p);
                             continue;
                         }
@@ -364,7 +348,7 @@ static void apc_cache_expunge(apc_cache_t* cache, size_t size)
 /* }}} */
 
 /* {{{ apc_cache_insert */
-int apc_cache_insert(apc_cache_t* cache,
+static inline int _apc_cache_insert(apc_cache_t* cache,
                      apc_cache_key_t key,
                      apc_cache_entry_t* value,
                      apc_context_t* ctxt,
@@ -380,7 +364,6 @@ int apc_cache_insert(apc_cache_t* cache,
     fprintf(stderr,"Inserting [%s]\n", value->data.file.filename);
 #endif
 
-    CACHE_LOCK(cache);
     process_pending_removals(cache);
 
     if(key.type == APC_CACHE_KEY_FILE) slot = &cache->slots[hash(key) % cache->num_slots];
@@ -395,14 +378,13 @@ int apc_cache_insert(apc_cache_t* cache,
                     remove_slot(cache, slot);
                     break;
                 }
-                CACHE_UNLOCK(cache);
                 return 0;
             } else if(cache->ttl && (*slot)->access_time < (t - cache->ttl)) {
                 remove_slot(cache, slot);
                 continue;
             }
         } else {   /* APC_CACHE_KEY_FPFILE */
-                if(!memcmp((*slot)->key.data.fpfile.fullpath, key.data.fpfile.fullpath, key.data.fpfile.fullpath_len+1)) {
+            if(!memcmp((*slot)->key.data.fpfile.fullpath, key.data.fpfile.fullpath, key.data.fpfile.fullpath_len+1)) {
                 /* Hrm.. it's already here, remove it and insert new one */
                 remove_slot(cache, slot);
                 break;
@@ -416,7 +398,6 @@ int apc_cache_insert(apc_cache_t* cache,
     }
 
     if ((*slot = make_slot(key, value, *slot, t)) == NULL) {
-        CACHE_UNLOCK(cache);
         return -1;
     }
 
@@ -425,27 +406,81 @@ int apc_cache_insert(apc_cache_t* cache,
     cache->header->num_entries++;
     cache->header->num_inserts++;
 
-    CACHE_UNLOCK(cache);
     return 1;
 }
 /* }}} */
 
+/* {{{ apc_cache_insert */
+int apc_cache_insert(apc_cache_t* cache, apc_cache_key_t key, apc_cache_entry_t* value, apc_context_t *ctxt, time_t t)
+{
+    int rval;
+    CACHE_LOCK(cache);
+    rval = _apc_cache_insert(cache, key, value, ctxt, t);
+    CACHE_UNLOCK(cache);
+    return rval;
+}
+/* }}} */
+
+/* {{{ apc_cache_insert */
+int *apc_cache_insert_mult(apc_cache_t* cache, apc_cache_key_t* keys, apc_cache_entry_t** values, apc_context_t *ctxt, time_t t, int num_entries)
+{
+    int *rval;
+    int i;
+
+    rval = emalloc(sizeof(int) * num_entries);
+    CACHE_LOCK(cache);
+    for (i=0; i < num_entries; i++) {
+        if (values[i]) {
+            ctxt->pool = values[i]->pool;
+            rval[i] = _apc_cache_insert(cache, keys[i], values[i], ctxt, t);
+        }
+    }
+    CACHE_UNLOCK(cache);
+    return rval;
+}
+/* }}} */
+
+
 /* {{{ apc_cache_user_insert */
 int apc_cache_user_insert(apc_cache_t* cache, apc_cache_key_t key, apc_cache_entry_t* value, apc_context_t* ctxt, time_t t, int exclusive TSRMLS_DC)
 {
     slot_t** slot;
-
+    unsigned int keylen = key.data.user.identifier_len+1;
+    unsigned int h = string_nhash_8(key.data.user.identifier, keylen);
+    apc_keyid_t *lastkey = &cache->header->lastkey;
+    
     if (!value) {
         return 0;
     }
+    
+    if(apc_cache_busy(cache)) {
+        /* cache cleanup in progress, do not wait */ 
+        return 0;
+    }
+
+    if(apc_cache_is_last_key(cache, &key, t)) {
+        /* potential cache slam */
+        return 0;
+    }
 
     CACHE_LOCK(cache);
-    process_pending_removals(cache);
 
-    slot = &cache->slots[string_nhash_8(key.data.user.identifier, key.data.user.identifier_len+1) % cache->num_slots];
+    memset(lastkey, 0, sizeof(apc_keyid_t));
+
+    lastkey->h = h;
+    lastkey->keylen = keylen;
+    lastkey->mtime = t;
+    
+    /* we do not reset lastkey after the insert. Whether it is inserted 
+     * or not, another insert in the same second is always a bad idea. 
+     */
+
+    process_pending_removals(cache);
+    
+    slot = &cache->slots[h % cache->num_slots];
 
     while (*slot) {
-        if (!memcmp((*slot)->key.data.user.identifier, key.data.user.identifier, key.data.user.identifier_len+1)) {
+        if (!memcmp((*slot)->key.data.user.identifier, key.data.user.identifier, keylen)) {
             /* 
              * At this point we have found the user cache entry.  If we are doing 
              * an exclusive insert (apc_add) we are going to bail right away if
@@ -453,10 +488,9 @@ int apc_cache_user_insert(apc_cache_t* cache, apc_cache_key_t key, apc_cache_ent
              * there is a ttl and the entry has not timed out yet.
              */
             if(exclusive && (  !(*slot)->value->data.user.ttl ||
-                              ( (*slot)->value->data.user.ttl && ((*slot)->creation_time + (*slot)->value->data.user.ttl) >= t ) 
+                              ( (*slot)->value->data.user.ttl && (time_t) ((*slot)->creation_time + (*slot)->value->data.user.ttl) >= t ) 
                             ) ) {
-                CACHE_UNLOCK(cache);
-                return 0;
+                goto fail;
             }
             remove_slot(cache, slot);
             break;
@@ -469,7 +503,7 @@ int apc_cache_user_insert(apc_cache_t* cache, apc_cache_key_t key, apc_cache_ent
          * we see if the entry has a hard ttl on it and remove it if it has been around longer than its ttl
          */
         if((cache->ttl && (*slot)->access_time < (t - cache->ttl)) || 
-           ((*slot)->value->data.user.ttl && ((*slot)->creation_time + (*slot)->value->data.user.ttl) < t)) {
+           ((*slot)->value->data.user.ttl && (time_t) ((*slot)->creation_time + (*slot)->value->data.user.ttl) < t)) {
             remove_slot(cache, slot);
             continue;
         }
@@ -477,10 +511,9 @@ int apc_cache_user_insert(apc_cache_t* cache, apc_cache_key_t key, apc_cache_ent
     }
 
     if ((*slot = make_slot(key, value, *slot, t)) == NULL) {
-        CACHE_UNLOCK(cache);
-        return 0;
-    }
-
+        goto fail;
+    } 
+    
     value->mem_size = ctxt->pool->size;
     cache->header->mem_size += ctxt->pool->size;
 
@@ -488,7 +521,13 @@ int apc_cache_user_insert(apc_cache_t* cache, apc_cache_key_t key, apc_cache_ent
     cache->header->num_inserts++;
 
     CACHE_UNLOCK(cache);
+
     return 1;
+
+fail:
+    CACHE_UNLOCK(cache);
+
+    return 0;
 }
 /* }}} */
 
@@ -570,7 +609,7 @@ apc_cache_entry_t* apc_cache_user_find(apc_cache_t* cache, char *strkey, int key
     while (*slot) {
         if (!memcmp((*slot)->key.data.user.identifier, strkey, keylen)) {
             /* Check to make sure this entry isn't expired by a hard TTL */
-            if((*slot)->value->data.user.ttl && ((*slot)->creation_time + (*slot)->value->data.user.ttl) < t) {
+            if((*slot)->value->data.user.ttl && (time_t) ((*slot)->creation_time + (*slot)->value->data.user.ttl) < t) {
                 remove_slot(cache, slot);
                 cache->header->num_misses++;
                 CACHE_UNLOCK(cache);
@@ -614,6 +653,7 @@ int _apc_cache_user_update(apc_cache_t* cache, char *strkey, int keylen, apc_cac
     while (*slot) {
         if (!memcmp((*slot)->key.data.user.identifier, strkey, keylen)) {
             retval = updater(cache, (*slot)->value, data);
+            (*slot)->key.mtime = apc_time();
             CACHE_UNLOCK(cache);
             return retval;
         }
@@ -656,7 +696,7 @@ int apc_cache_delete(apc_cache_t* cache, char *filename, int filename_len)
 
     TSRMLS_FETCH();
 
-    t = sapi_get_request_time(TSRMLS_C);
+    t = apc_time();
 
     /* try to create a cache key; if we fail, give up on caching */
     if (!apc_cache_make_file_key(&key, filename, PG(include_path), t TSRMLS_CC)) {
@@ -784,7 +824,7 @@ int apc_cache_make_file_key(apc_cache_key_t* key,
      * tiny safety is easier than educating the world.  This is now
      * configurable, but the default is still 2 seconds.
      */
-    if(APCG(file_update_protection) && (t - fileinfo.st_buf.sb.st_mtime < APCG(file_update_protection))) { 
+    if(APCG(file_update_protection) && (t - fileinfo.st_buf.sb.st_mtime < APCG(file_update_protection)) && !APCG(force_file_update)) {
 #ifdef __DEBUG_APC__
         fprintf(stderr,"File is too new %s (%d - %d) - bailing\n",filename,t,fileinfo.st_buf.sb.st_mtime);
 #endif
@@ -856,6 +896,9 @@ apc_cache_entry_t* apc_cache_make_file_entry(const char* filename,
     entry->data.file.op_array  = op_array;
     entry->data.file.functions = functions;
     entry->data.file.classes   = classes;
+
+    entry->data.file.halt_offset = apc_file_halt_offset(filename);
+
     entry->type = APC_CACHE_ENTRY_FILE;
     entry->ref_count = 0;
     entry->mem_size = 0;
@@ -867,21 +910,18 @@ apc_cache_entry_t* apc_cache_make_file_entry(const char* filename,
 /* {{{ apc_cache_store_zval */
 zval* apc_cache_store_zval(zval* dst, const zval* src, apc_context_t* ctxt)
 {
-    /* Maintain a list of zvals we've copied to properly handle recursive structures */
-    HashTable *old;
     TSRMLS_FETCH();
-    old = APCG(copied_zvals);
-    APCG(copied_zvals) = emalloc(sizeof(HashTable));
-    zend_hash_init(APCG(copied_zvals), 0, NULL, NULL, 0);
 
-    dst = apc_copy_zval(dst, src, ctxt);
-
-    if(APCG(copied_zvals)) {
-        zend_hash_destroy(APCG(copied_zvals));
-        efree(APCG(copied_zvals));
+    if (Z_TYPE_P(src) == IS_ARRAY) {
+        /* Maintain a list of zvals we've copied to properly handle recursive structures */
+        zend_hash_init(&APCG(copied_zvals), 0, NULL, NULL, 0);
+        dst = apc_copy_zval(dst, src, ctxt);
+        zend_hash_destroy(&APCG(copied_zvals));
+        APCG(copied_zvals).nTableSize=0;
+    } else {
+        dst = apc_copy_zval(dst, src, ctxt);
     }
 
-    APCG(copied_zvals) = old;
 
     return dst;
 }
@@ -890,22 +930,18 @@ zval* apc_cache_store_zval(zval* dst, const zval* src, apc_context_t* ctxt)
 /* {{{ apc_cache_fetch_zval */
 zval* apc_cache_fetch_zval(zval* dst, const zval* src, apc_context_t* ctxt)
 {
-    /* Maintain a list of zvals we've copied to properly handle recursive structures */
-    HashTable *old;
     TSRMLS_FETCH();
 
-    old = APCG(copied_zvals);
-    APCG(copied_zvals) = emalloc(sizeof(HashTable));
-    zend_hash_init(APCG(copied_zvals), 0, NULL, NULL, 0);
-
-    dst = apc_copy_zval(dst, src, ctxt);
-
-    if(APCG(copied_zvals)) {
-        zend_hash_destroy(APCG(copied_zvals));
-        efree(APCG(copied_zvals));
+    if (Z_TYPE_P(src) == IS_ARRAY) {
+        /* Maintain a list of zvals we've copied to properly handle recursive structures */
+        zend_hash_init(&APCG(copied_zvals), 0, NULL, NULL, 0);
+        dst = apc_copy_zval(dst, src, ctxt);
+        zend_hash_destroy(&APCG(copied_zvals));
+        APCG(copied_zvals).nTableSize=0;
+    } else {
+        dst = apc_copy_zval(dst, src, ctxt);
     }
 
-    APCG(copied_zvals) = old;
 
     return dst;
 }
@@ -1101,6 +1137,26 @@ zend_bool apc_cache_busy(apc_cache_t* cache)
 }
 /* }}} */
 
+/* {{{ apc_cache_is_last_key */
+zend_bool apc_cache_is_last_key(apc_cache_t* cache, apc_cache_key_t* key, time_t t)
+{
+    apc_keyid_t *lastkey = &cache->header->lastkey;
+    unsigned int keylen = key->data.user.identifier_len+1;
+    unsigned int h = string_nhash_8(key->data.user.identifier, keylen);
+
+    /* unlocked reads, but we're not shooting for 100% success with this */
+    if(lastkey->h == h && keylen == lastkey->keylen) {
+        if(lastkey->mtime == t) {
+            /* potential cache slam */
+            apc_wprint("Potential cache slam averted for key '%s'", key->data.user.identifier);
+            return 1;
+        }
+    }
+
+    return 0;
+}
+/* }}} */
+
 #if NONBLOCKING_LOCK_AVAILABLE
 /* {{{ apc_cache_write_lock */
 zend_bool apc_cache_write_lock(apc_cache_t* cache)
