@@ -26,7 +26,7 @@
 
  */
 
-/* $Id: apc_cache.h,v 3.59 2008/11/24 21:34:47 shire Exp $ */
+/* $Id: apc_cache.h 281919 2009-06-10 11:23:55Z gopalv $ */
 
 #ifndef APC_CACHE_H
 #define APC_CACHE_H
@@ -82,6 +82,15 @@ struct apc_cache_key_t {
     unsigned char type;
     unsigned char md5[16];        /* md5 hash of the source file */
 };
+
+
+typedef struct apc_keyid_t apc_keyid_t;
+
+struct apc_keyid_t {
+    unsigned int h;
+    unsigned int keylen;
+    time_t mtime;
+};
 /* }}} */
 
 /* {{{ struct definition: apc_cache_entry_t */
@@ -91,6 +100,7 @@ typedef union _apc_cache_entry_value_t {
         zend_op_array* op_array;    /* op_array allocated in shared memory */
         apc_function_t* functions;  /* array of apc_function_t's */
         apc_class_t* classes;       /* array of apc_class_t's */
+        long halt_offset;           /* value of __COMPILER_HALT_OFFSET__ for the file */
     } file;
     struct {
         char *info;
@@ -161,6 +171,9 @@ extern int apc_cache_insert(T cache, apc_cache_key_t key,
 extern int apc_cache_user_insert(T cache, apc_cache_key_t key,
                             apc_cache_entry_t* value, apc_context_t* ctxt, time_t t, int exclusive TSRMLS_DC);
 
+extern int *apc_cache_insert_mult(apc_cache_t* cache, apc_cache_key_t* keys,
+                            apc_cache_entry_t** values, apc_context_t *ctxt, time_t t, int num_entries);
+
 /*
  * apc_cache_find searches for a cache entry by filename, and returns a
  * pointer to the entry if found, NULL otherwise.
@@ -230,6 +243,10 @@ extern apc_cache_entry_t* apc_cache_make_file_entry(const char* filename,
                                                     apc_function_t* functions,
                                                     apc_class_t* classes,
                                                     apc_context_t* ctxt);
+
+
+zend_bool apc_compile_cache_entry(apc_cache_key_t key, zend_file_handle* h, int type, time_t t, zend_op_array** op_array_pp, apc_cache_entry_t** cache_entry_pp TSRMLS_DC);
+
 /*
  * apc_cache_make_user_entry creates an apc_cache_entry_t object given an info string
  * and the zval to be stored.
@@ -258,7 +275,7 @@ typedef struct apc_cache_link_t apc_cache_link_t;
 struct apc_cache_link_t {
     apc_cache_link_data_t data;
     unsigned char type;
-    int num_hits;
+    unsigned long num_hits;
     time_t mtime;
     time_t creation_time;
     time_t deletion_time;
@@ -273,15 +290,15 @@ struct apc_cache_link_t {
 typedef struct apc_cache_info_t apc_cache_info_t;
 struct apc_cache_info_t {
     int num_slots;
-    int num_hits;
-    int num_misses;
+    unsigned long num_hits;
+    unsigned long num_misses;
+    unsigned long num_inserts;
+    unsigned long expunges;
     int ttl;
     apc_cache_link_t* list;
     apc_cache_link_t* deleted_list;
     time_t start_time;
-    int expunges;
     int num_entries;
-    int num_inserts;
     size_t mem_size;
 };
 /* }}} */
@@ -292,7 +309,7 @@ struct slot_t {
     apc_cache_key_t key;        /* slot key */
     apc_cache_entry_t* value;   /* slot value */
     slot_t* next;               /* next slot in linked list */
-    int num_hits;               /* number of hits to this bucket */
+    unsigned long num_hits;     /* number of hits to this bucket */
     time_t creation_time;       /* time slot was initialized */
     time_t deletion_time;       /* time slot was removed from cache */
     time_t access_time;         /* time slot was last accessed */
@@ -305,15 +322,16 @@ typedef struct cache_header_t cache_header_t;
 struct cache_header_t {
     apc_lck_t lock;             /* read/write lock (exclusive blocking cache lock) */
     apc_lck_t wrlock;           /* write lock (non-blocking used to prevent cache slams) */
-    int num_hits;               /* total successful hits in cache */
-    int num_misses;             /* total unsuccessful hits in cache */
-    int num_inserts;            /* total successful inserts in cache */
+    unsigned long num_hits;     /* total successful hits in cache */
+    unsigned long num_misses;   /* total unsuccessful hits in cache */
+    unsigned long num_inserts;  /* total successful inserts in cache */
+    unsigned long expunges;     /* total number of expunges */
     slot_t* deleted_list;       /* linked list of to-be-deleted slots */
     time_t start_time;          /* time the above counters were reset */
-    int expunges;               /* total number of expunges */
     zend_bool busy;             /* Flag to tell clients when we are busy cleaning the cache */
     int num_entries;            /* Statistic on the number of entries */
     size_t mem_size;            /* Statistic on the memory size used by this cache */
+    apc_keyid_t lastkey;        /* the key that is being inserted (user cache) */
 };
 /* }}} */
 
@@ -338,6 +356,7 @@ extern void apc_cache_unlock(apc_cache_t* cache);
 extern zend_bool apc_cache_busy(apc_cache_t* cache);
 extern zend_bool apc_cache_write_lock(apc_cache_t* cache);
 extern void apc_cache_write_unlock(apc_cache_t* cache);
+extern zend_bool apc_cache_is_last_key(apc_cache_t* cache, apc_cache_key_t* key, time_t t);
 
 /* used by apc_rfc1867 to update data in-place - not to be used elsewhere */
 
