@@ -28,7 +28,7 @@
 
  */
 
-/* $Id: apc_main.c,v 3.134 2008/11/26 07:48:39 gopalv Exp $ */
+/* $Id: apc_main.c 284282 2009-07-17 21:50:33Z shire $ */
 
 #include "apc_php.h"
 #include "apc_main.h"
@@ -65,21 +65,28 @@ static zend_compile_t* set_compile_hook(zend_compile_t *ptr)
 /* }}} */
 
 /* {{{ install_function */
-static int install_function(apc_function_t fn, apc_context_t* ctxt TSRMLS_DC)
+static int install_function(apc_function_t fn, apc_context_t* ctxt, int lazy TSRMLS_DC)
 {
-    zend_function *func;
     int status;
 
-    func = apc_copy_function_for_execution(fn.function, ctxt);
 
-    status =  zend_hash_add(EG(function_table),
-                      fn.name,
-                      fn.name_len+1,
-                      func,
-                      sizeof(fn.function[0]),
-                      NULL);
-
-    efree(func);
+    if(lazy && fn.name[0] != '\0' && strncmp(fn.name, "__autoload", fn.name_len) != 0) {
+        status = zend_hash_add(APCG(lazy_function_table),
+                              fn.name,
+                              fn.name_len+1,
+                              &fn,
+                              sizeof(apc_function_t),
+                              NULL);
+    } else {
+        zend_function *func = apc_copy_function_for_execution(fn.function, ctxt);
+        status = zend_hash_add(EG(function_table),
+                              fn.name,
+                              fn.name_len+1,
+                              func,
+                              sizeof(fn.function[0]),
+                              NULL);
+        efree(func);
+    }
 
     if (status == FAILURE) {
         /* apc_eprint("Cannot redeclare %s()", fn.name); */
@@ -89,15 +96,38 @@ static int install_function(apc_function_t fn, apc_context_t* ctxt TSRMLS_DC)
 }
 /* }}} */
 
+/* {{{ apc_lookup_function_hook */
+int apc_lookup_function_hook(char *name, int len, ulong hash, zend_function **fe) {
+    apc_function_t *fn;
+    int status = FAILURE;
+    apc_context_t ctxt = {0,};
+    TSRMLS_FETCH();
+
+    ctxt.pool = apc_pool_create(APC_UNPOOL, apc_php_malloc, apc_php_free, apc_sma_protect, apc_sma_unprotect);
+    ctxt.copy = APC_COPY_OUT_OPCODE;
+
+    if(zend_hash_quick_find(APCG(lazy_function_table), name, len, hash, (void**)&fn) == SUCCESS) {
+        *fe = apc_copy_function_for_execution(fn->function, &ctxt);
+        status = zend_hash_add(EG(function_table),
+                                  fn->name,
+                                  fn->name_len+1,
+                                  *fe,
+                                  sizeof(zend_function),
+                                  NULL);
+    }
+
+    return status;
+}
+/* }}} */
+
 /* {{{ install_class */
-static int install_class(apc_class_t cl, apc_context_t* ctxt TSRMLS_DC)
+static int install_class(apc_class_t cl, apc_context_t* ctxt, int lazy TSRMLS_DC)
 {
     zend_class_entry* class_entry = cl.class_entry;
     zend_class_entry* parent = NULL;
     int status;
     zend_class_entry** allocated_ce = NULL;
 
-
     /* Special case for mangled names. Mangled names are unique to a file.
      * There is no way two classes with the same mangled name will occur,
      * unless a file is included twice. And if in case, a file is included
@@ -112,6 +142,19 @@ static int install_class(apc_class_t cl, apc_context_t* ctxt TSRMLS_DC)
         }
     }
 
+    if(lazy && cl.name_len != 0 && cl.name[0] != '\0') {
+        status = zend_hash_add(APCG(lazy_class_table),
+                               cl.name,
+                               cl.name_len+1,
+                               &cl,
+                               sizeof(apc_class_t),
+                               NULL);
+        if(status == FAILURE) {
+            zend_error(E_ERROR, "Cannot redeclare class %s", cl.name);
+        }
+        return status;
+    }
+
     /*
      * XXX: We need to free this somewhere...
      */
@@ -178,6 +221,37 @@ static int install_class(apc_class_t cl, apc_context_t* ctxt TSRMLS_DC)
 }
 /* }}} */
 
+/* {{{ apc_lookup_class_hook */
+int apc_lookup_class_hook(char *name, int len, ulong hash, zend_class_entry ***ce) {
+
+    apc_class_t *cl;
+    apc_context_t ctxt = {0,};
+    TSRMLS_FETCH();
+
+    if(zend_is_compiling(TSRMLS_C)) { return FAILURE; }
+
+    if(zend_hash_quick_find(APCG(lazy_class_table), name, len, hash, (void**)&cl) == FAILURE) {
+        return FAILURE;
+    }
+
+    ctxt.pool = apc_pool_create(APC_UNPOOL, apc_php_malloc, apc_php_free, apc_sma_protect, apc_sma_unprotect);
+    ctxt.copy = APC_COPY_OUT_OPCODE;
+
+    if(install_class(*cl, &ctxt, 0 TSRMLS_CC) == FAILURE) {
+        apc_wprint("apc_lookup_class_hook: could not install %s", name);
+        return FAILURE;
+    }
+
+    if(zend_hash_quick_find(EG(class_table), name, len, hash, (void**)ce) == FAILURE) {
+        apc_wprint("apc_lookup_class_hook: known error trying to fetch class %s", name);
+        return FAILURE;
+    }
+
+    return SUCCESS;
+
+}
+/* }}} */
+
 /* {{{ uninstall_class */
 static int uninstall_class(apc_class_t cl TSRMLS_DC)
 {
@@ -193,6 +267,61 @@ static int uninstall_class(apc_class_t cl TSRMLS_DC)
 }
 /* }}} */
 
+/* {{{ copy_function_name (taken from zend_builtin_functions.c to ensure future compatibility with APC) */
+static int copy_function_name(apc_function_t *pf TSRMLS_DC, int num_args, va_list args, zend_hash_key *hash_key)
+{
+    zval *internal_ar = va_arg(args, zval *),
+         *user_ar     = va_arg(args, zval *);
+    zend_function *func = pf->function;
+
+    if (hash_key->nKeyLength == 0 || hash_key->arKey[0] == 0) {
+        return 0;
+    }
+
+    if (func->type == ZEND_INTERNAL_FUNCTION) {
+        add_next_index_stringl(internal_ar, hash_key->arKey, hash_key->nKeyLength-1, 1);
+    } else if (func->type == ZEND_USER_FUNCTION) {
+        add_next_index_stringl(user_ar, hash_key->arKey, hash_key->nKeyLength-1, 1);
+    }
+
+    return 0;
+}
+
+/* {{{ copy_class_or_interface_name (taken from zend_builtin_functions.c to ensure future compatibility with APC) */
+static int copy_class_or_interface_name(apc_class_t *cl TSRMLS_DC, int num_args, va_list args, zend_hash_key *hash_key)
+{
+    zval *array = va_arg(args, zval *);
+    zend_uint mask = va_arg(args, zend_uint);
+    zend_uint comply = va_arg(args, zend_uint);
+    zend_uint comply_mask = (comply)? mask:0;
+    zend_class_entry *ce  = cl->class_entry;
+
+    if ((hash_key->nKeyLength==0 || hash_key->arKey[0]!=0)
+        && (comply_mask == (ce->ce_flags & mask))) {
+        add_next_index_stringl(array, ce->name, ce->name_length, 1);
+    }
+    return ZEND_HASH_APPLY_KEEP;
+}
+/* }}} */
+
+/* }}} */
+
+/* {{{ apc_defined_function_hook */
+int apc_defined_function_hook(zval *internal, zval *user) {
+    TSRMLS_FETCH();
+  zend_hash_apply_with_arguments(APCG(lazy_function_table) TSRMLS_CC, (apply_func_args_t) copy_function_name, 2, internal, user);
+  return 1;
+}
+/* }}} */
+
+/* {{{ apc_declared_class_hook */
+int apc_declared_class_hook(zval *classes, zend_uint mask, zend_uint comply) {
+    TSRMLS_FETCH();
+  zend_hash_apply_with_arguments(APCG(lazy_class_table) TSRMLS_CC, (apply_func_args_t) copy_class_or_interface_name, 3, classes, mask, comply);
+  return 1;
+}
+/* }}} */
+
 /* {{{ compare_file_handles */
 static int compare_file_handles(void* a, void* b)
 {
@@ -217,7 +346,7 @@ static zend_op_array* cached_compile(zend_file_handle* h,
 
     if (cache_entry->data.file.classes) {
         for (i = 0; cache_entry->data.file.classes[i].class_entry != NULL; i++) {
-            if(install_class(cache_entry->data.file.classes[i], ctxt TSRMLS_CC) == FAILURE) {
+            if(install_class(cache_entry->data.file.classes[i], ctxt, APCG(lazy_classes) TSRMLS_CC) == FAILURE) {
                 goto default_compile;
             }
         }
@@ -225,10 +354,12 @@ static zend_op_array* cached_compile(zend_file_handle* h,
 
     if (cache_entry->data.file.functions) {
         for (i = 0; cache_entry->data.file.functions[i].function != NULL; i++) {
-            install_function(cache_entry->data.file.functions[i], ctxt TSRMLS_CC);
+            install_function(cache_entry->data.file.functions[i], ctxt, APCG(lazy_functions) TSRMLS_CC);
         }
     }
 
+    apc_do_halt_compiler_register(cache_entry->data.file.filename, cache_entry->data.file.halt_offset TSRMLS_CC);
+
 
     return apc_copy_op_array_for_execution(NULL, cache_entry->data.file.op_array, ctxt TSRMLS_CC);
 
@@ -264,137 +395,34 @@ default_compile:
 }
 /* }}} */
 
-/* {{{ my_compile_file
-   Overrides zend_compile_file */
-static zend_op_array* my_compile_file(zend_file_handle* h,
-                                               int type TSRMLS_DC)
-{
-    apc_cache_key_t key;
-    apc_cache_entry_t* cache_entry;
-    zend_op_array* op_array;
-    int num_functions, num_classes, ret;
-    zend_op_array* alloc_op_array;
+/* {{{ apc_compile_cache_entry  */
+zend_bool apc_compile_cache_entry(apc_cache_key_t key, zend_file_handle* h, int type, time_t t, zend_op_array** op_array, apc_cache_entry_t** cache_entry TSRMLS_DC) {
+    int num_functions, num_classes;
     apc_function_t* alloc_functions;
+    zend_op_array* alloc_op_array;
     apc_class_t* alloc_classes;
-    time_t t;
     char *path;
-    apc_context_t ctxt = {0,};
-
-    if (!APCG(enabled) || apc_cache_busy(apc_cache)) {
-        return old_compile_file(h, type TSRMLS_CC);
-    }
-
-    /* check our regular expression filters */
-    if (APCG(filters) && apc_compiled_filters) {
-        int ret = apc_regex_match_array(apc_compiled_filters, h->filename);
-        if(ret == APC_NEGATIVE_MATCH || (ret != APC_POSITIVE_MATCH && !APCG(cache_by_default))) {
-            return old_compile_file(h, type TSRMLS_CC);
-        }
-    } else if(!APCG(cache_by_default)) {
-        return old_compile_file(h, type TSRMLS_CC);
-    }
-
-
-    t = sapi_get_request_time(TSRMLS_C);
-
-#ifdef __DEBUG_APC__
-    fprintf(stderr,"1. h->opened_path=[%s]  h->filename=[%s]\n", h->opened_path?h->opened_path:"null",h->filename);
-#endif
-
-    /* try to create a cache key; if we fail, give up on caching */
-    if (!apc_cache_make_file_key(&key, h->filename, PG(include_path), t TSRMLS_CC)) {
-        return old_compile_file(h, type TSRMLS_CC);
-    }
-
-
-    if(!APCG(force_file_update)) {
-        /* search for the file in the cache */
-        cache_entry = apc_cache_find(apc_cache, key, t);
-        ctxt.force_update = 0;
-    } else {
-        cache_entry = NULL;
-        ctxt.force_update = 1;
-    }
-
-    if (cache_entry != NULL) {
-        int dummy = 1;
-
-        ctxt.pool = apc_pool_create(APC_UNPOOL, apc_php_malloc, apc_php_free);
-        ctxt.copy = APC_COPY_OUT_OPCODE;
-
-        if (h->opened_path == NULL) {
-            h->opened_path = estrdup(cache_entry->data.file.filename);
-        }
-        zend_hash_add(&EG(included_files), h->opened_path, strlen(h->opened_path)+1, (void *)&dummy, sizeof(int), NULL);
-
-        zend_llist_add_element(&CG(open_files), h); /* We leak fds without this hack */
-
-        apc_stack_push(APCG(cache_stack), cache_entry);
-        op_array = cached_compile(h, type, &ctxt TSRMLS_CC);
-        if(op_array) {
-#ifdef APC_FILEHITS
-            /* If the file comes from the cache, add it to the global request file list */
-            add_next_index_string(APCG(filehits), h->filename, 1);
-#endif
-            /* this is an unpool, which has no cleanup - this only free's the pool header */
-            apc_pool_destroy(ctxt.pool);
-            return op_array;
-        }
-        if(APCG(report_autofilter)) {
-            apc_wprint("Recompiling %s", cache_entry->data.file.filename);
-        }
-        /* TODO: check what happens with EG(included_files) */
-    }
+    apc_context_t ctxt;
 
     /* remember how many functions and classes existed before compilation */
     num_functions = zend_hash_num_elements(CG(function_table));
     num_classes   = zend_hash_num_elements(CG(class_table));
 
-    /* compile the file using the default compile function */
-    op_array = old_compile_file(h, type TSRMLS_CC);
-    if (op_array == NULL) {
-        return NULL;
+    /* compile the file using the default compile function,  *
+     * we set *op_array here so we return opcodes during     *
+     * a failure.  We should not return prior to this line.  */
+    *op_array = old_compile_file(h, type TSRMLS_CC);
+    if (*op_array == NULL) {
+        return FAILURE;
     }
 
-    /* check our regular expression filters */
-    if (APCG(filters) && apc_compiled_filters && h->opened_path) {
-        int ret = apc_regex_match_array(apc_compiled_filters, h->opened_path);
-        if(ret == APC_NEGATIVE_MATCH || (ret != APC_POSITIVE_MATCH && !APCG(cache_by_default))) {
-            /* never cache, never find */
-            return op_array;
-        }
+    ctxt.pool = apc_pool_create(APC_MEDIUM_POOL, apc_sma_malloc, apc_sma_free, 
+                                                 apc_sma_protect, apc_sma_unprotect);
+    if (!ctxt.pool) {
+        apc_wprint("Unable to allocate memory for pool.");
+        return FAILURE;
     }
-
-    /* Make sure the mtime reflects the files last known mtime in the case of fpstat==0 */
-    if(key.type == APC_CACHE_KEY_FPFILE) {
-        apc_fileinfo_t fileinfo;
-        struct stat *tmp_buf = NULL;
-        if(!strcmp(SG(request_info).path_translated, h->filename)) {
-            tmp_buf = sapi_get_stat(TSRMLS_C);  /* Apache has already done this stat() for us */
-        }
-        if(tmp_buf) {
-            fileinfo.st_buf.sb = *tmp_buf;
-        } else {
-            if (apc_search_paths(h->filename, PG(include_path), &fileinfo) != 0) {
-#ifdef __DEBUG_APC__
-                fprintf(stderr,"Stat failed %s - bailing (%s) (%d)\n",filename,SG(request_info).path_translated);
-#endif
-                return op_array;
-            }
-        }
-        key.mtime = fileinfo.st_buf.sb.st_mtime;
-    }
-
-    HANDLE_BLOCK_INTERRUPTIONS();
-
-#if NONBLOCKING_LOCK_AVAILABLE
-    if(APCG(write_lock)) {
-        if(!apc_cache_write_lock(apc_cache)) {
-            HANDLE_UNBLOCK_INTERRUPTIONS();
-            return op_array;
-        }
-    }
-#endif
+    ctxt.copy = APC_COPY_IN_OPCODE;
 
     if(APCG(file_md5)) {
         int n;
@@ -424,18 +452,14 @@ static zend_op_array* my_compile_file(zend_file_handle* h,
         }
     }
 
-    APCG(current_cache) = apc_cache;
-    ctxt.pool = apc_pool_create(APC_MEDIUM_POOL, apc_sma_malloc, apc_sma_free);
-    ctxt.copy = APC_COPY_IN_OPCODE;
-
-    if(!(alloc_op_array = apc_copy_op_array(NULL, op_array, &ctxt TSRMLS_CC))) {
+    if(!(alloc_op_array = apc_copy_op_array(NULL, *op_array, &ctxt TSRMLS_CC))) {
         goto freepool;
     }
 
     if(!(alloc_functions = apc_copy_new_functions(num_functions, &ctxt TSRMLS_CC))) {
         goto freepool;
     }
-    if(!(alloc_classes = apc_copy_new_classes(op_array, num_classes, &ctxt TSRMLS_CC))) {
+    if(!(alloc_classes = apc_copy_new_classes(*op_array, num_classes, &ctxt TSRMLS_CC))) {
         goto freepool;
     }
 
@@ -446,16 +470,150 @@ static zend_op_array* my_compile_file(zend_file_handle* h,
     fprintf(stderr,"2. h->opened_path=[%s]  h->filename=[%s]\n", h->opened_path?h->opened_path:"null",h->filename);
 #endif
 
-    if(!(cache_entry = apc_cache_make_file_entry(path, alloc_op_array, alloc_functions, alloc_classes, &ctxt))) {
+    if(!(*cache_entry = apc_cache_make_file_entry(path, alloc_op_array, alloc_functions, alloc_classes, &ctxt))) {
         goto freepool;
     }
 
-    if ((ret = apc_cache_insert(apc_cache, key, cache_entry, &ctxt, t)) != 1) {
+    return SUCCESS;
+
 freepool:
-        apc_pool_destroy(ctxt.pool);
-        ctxt.pool = NULL;
+    apc_pool_destroy(ctxt.pool);
+    ctxt.pool = NULL;
+
+    return FAILURE;
+
+}
+/* }}} */
+
+/* {{{ my_compile_file
+   Overrides zend_compile_file */
+static zend_op_array* my_compile_file(zend_file_handle* h,
+                                               int type TSRMLS_DC)
+{
+    apc_cache_key_t key;
+    apc_cache_entry_t* cache_entry;
+    zend_op_array* op_array = NULL;
+    time_t t;
+    apc_context_t ctxt = {0,};
+    int bailout=0;
+
+    if (!APCG(enabled) || apc_cache_busy(apc_cache)) {
+        return old_compile_file(h, type TSRMLS_CC);
     }
 
+    /* check our regular expression filters */
+    if (APCG(filters) && APCG(compiled_filters) && h->opened_path) {
+        int ret = apc_regex_match_array(APCG(compiled_filters), h->opened_path);
+
+        if(ret == APC_NEGATIVE_MATCH || (ret != APC_POSITIVE_MATCH && !APCG(cache_by_default))) {
+            return old_compile_file(h, type TSRMLS_CC);
+        }
+    } else if(!APCG(cache_by_default)) {
+        return old_compile_file(h, type TSRMLS_CC);
+    }
+    APCG(current_cache) = apc_cache;
+
+
+    t = apc_time();
+
+#ifdef __DEBUG_APC__
+    fprintf(stderr,"1. h->opened_path=[%s]  h->filename=[%s]\n", h->opened_path?h->opened_path:"null",h->filename);
+#endif
+
+    /* try to create a cache key; if we fail, give up on caching */
+    if (!apc_cache_make_file_key(&key, h->filename, PG(include_path), t TSRMLS_CC)) {
+        return old_compile_file(h, type TSRMLS_CC);
+    }
+
+
+    if(!APCG(force_file_update)) {
+        /* search for the file in the cache */
+        cache_entry = apc_cache_find(apc_cache, key, t);
+        ctxt.force_update = 0;
+    } else {
+        cache_entry = NULL;
+        ctxt.force_update = 1;
+    }
+
+    if (cache_entry != NULL) {
+        int dummy = 1;
+
+        ctxt.pool = apc_pool_create(APC_UNPOOL, apc_php_malloc, apc_php_free,
+                                                apc_sma_protect, apc_sma_unprotect);
+        if (!ctxt.pool) {
+            apc_wprint("Unable to allocate memory for pool.");
+            return old_compile_file(h, type TSRMLS_CC);
+        }
+        ctxt.copy = APC_COPY_OUT_OPCODE;
+
+        if (h->opened_path == NULL) {
+            h->opened_path = estrdup(cache_entry->data.file.filename);
+        }
+        zend_hash_add(&EG(included_files), h->opened_path, strlen(h->opened_path)+1, (void *)&dummy, sizeof(int), NULL);
+
+        zend_llist_add_element(&CG(open_files), h); /* We leak fds without this hack */
+
+        apc_stack_push(APCG(cache_stack), cache_entry);
+        op_array = cached_compile(h, type, &ctxt TSRMLS_CC);
+        if(op_array) {
+#ifdef APC_FILEHITS
+            /* If the file comes from the cache, add it to the global request file list */
+            add_next_index_string(APCG(filehits), h->filename, 1);
+#endif
+            /* this is an unpool, which has no cleanup - this only free's the pool header */
+            apc_pool_destroy(ctxt.pool);
+            return op_array;
+        }
+        if(APCG(report_autofilter)) {
+            apc_wprint("Recompiling %s", cache_entry->data.file.filename);
+        }
+        /* TODO: check what happens with EG(included_files) */
+    }
+
+    /* Make sure the mtime reflects the files last known mtime in the case of fpstat==0 */
+    if(key.type == APC_CACHE_KEY_FPFILE) {
+        apc_fileinfo_t fileinfo;
+        struct stat *tmp_buf = NULL;
+        if(!strcmp(SG(request_info).path_translated, h->filename)) {
+            tmp_buf = sapi_get_stat(TSRMLS_C);  /* Apache has already done this stat() for us */
+        }
+        if(tmp_buf) {
+            fileinfo.st_buf.sb = *tmp_buf;
+        } else {
+            if (apc_search_paths(h->filename, PG(include_path), &fileinfo) != 0) {
+#ifdef __DEBUG_APC__
+                fprintf(stderr,"Stat failed %s - bailing (%s) (%d)\n",h->filename,SG(request_info).path_translated);
+#endif
+                return op_array;
+            }
+        }
+        key.mtime = fileinfo.st_buf.sb.st_mtime;
+    }
+
+    HANDLE_BLOCK_INTERRUPTIONS();
+
+#if NONBLOCKING_LOCK_AVAILABLE
+    if(APCG(write_lock)) {
+        if(!apc_cache_write_lock(apc_cache)) {
+            HANDLE_UNBLOCK_INTERRUPTIONS();
+            return old_compile_file(h, type TSRMLS_CC);
+        }
+    }
+#endif
+
+    zend_try {
+        if (apc_compile_cache_entry(key, h, type, t, &op_array, &cache_entry TSRMLS_CC) == SUCCESS) {
+            ctxt.pool = cache_entry->pool;
+            ctxt.copy = APC_COPY_IN_OPCODE;
+            if (apc_cache_insert(apc_cache, key, cache_entry, &ctxt, t) != 1) {
+                apc_pool_destroy(ctxt.pool);
+                ctxt.pool = NULL;
+            }
+        }
+    } zend_catch {
+        bailout=1; /* in the event of a bailout, ensure we don't create a dead-lock */
+    } zend_end_try();
+
     APCG(current_cache) = NULL;
 
 #if NONBLOCKING_LOCK_AVAILABLE
@@ -465,6 +623,8 @@ freepool:
 #endif
     HANDLE_UNBLOCK_INTERRUPTIONS();
 
+    if (bailout) zend_bailout();
+
     return op_array;
 }
 /* }}} */
@@ -493,14 +653,14 @@ static zval* data_unserialize(const char *filename)
 
     tmp = contents = malloc(len);
 
-	if(!contents) {
-		return NULL;
-	}
+    if(!contents) {
+       return NULL;
+    }
 
     if(fread(contents, 1, len, fp) < 1) {	
-		free(contents);
-		return NULL;
-	}
+      free(contents);
+      return NULL;
+    }
 
     MAKE_STD_ZVAL(retval);
 
@@ -601,8 +761,6 @@ int apc_module_init(int module_number TSRMLS_DC)
     apc_cache = apc_cache_create(APCG(num_files_hint), APCG(gc_ttl), APCG(ttl));
     apc_user_cache = apc_cache_create(APCG(user_entries_hint), APCG(gc_ttl), APCG(user_ttl));
 
-    apc_compiled_filters = apc_regex_compile_array(APCG(filters) TSRMLS_CC);
-
     /* override compilation */
     old_compile_file = zend_compile_file;
     zend_compile_file = my_compile_file;
@@ -613,6 +771,22 @@ int apc_module_init(int module_number TSRMLS_DC)
 
     apc_data_preload(TSRMLS_C);
 
+#if APC_HAVE_LOOKUP_HOOKS
+    if(APCG(lazy_functions)) {
+        zend_set_lookup_function_hook(apc_lookup_function_hook TSRMLS_CC);
+        zend_set_defined_function_hook(apc_defined_function_hook TSRMLS_CC);
+    }
+    if(APCG(lazy_classes)) {
+        zend_set_lookup_class_hook(apc_lookup_class_hook TSRMLS_CC);
+        zend_set_declared_class_hook(apc_declared_class_hook TSRMLS_CC);
+    }
+#else
+    if(APCG(lazy_functions) || APCG(lazy_classes)) {
+        apc_wprint("Lazy function/class loading not available with this version of PHP, please disable APC lazy loading.");
+        APCG(lazy_functions) = APCG(lazy_classes) = 0;
+    }
+#endif
+
     APCG(initialized) = 1;
     return 0;
 }
@@ -677,6 +851,7 @@ int apc_process_shutdown(TSRMLS_D)
 }
 /* }}} */
 
+
 /* {{{ apc_deactivate */
 static void apc_deactivate(TSRMLS_D)
 {
@@ -707,6 +882,7 @@ static void apc_deactivate(TSRMLS_D)
                 }
 
                 zce = *pzce;
+
                 zend_hash_del(EG(class_table),
                     cache_entry->data.file.classes[i].name,
                     cache_entry->data.file.classes[i].name_len+1);
@@ -724,7 +900,20 @@ static void apc_deactivate(TSRMLS_D)
 int apc_request_init(TSRMLS_D)
 {
     apc_stack_clear(APCG(cache_stack));
-    APCG(copied_zvals) = NULL;
+    if (!APCG(compiled_filters) && APCG(filters)) {
+        /* compile regex filters here to avoid race condition between MINIT of PCRE and APC.
+         * This should be moved to apc_cache_create() if this race condition between modules is resolved */
+        APCG(compiled_filters) = apc_regex_compile_array(APCG(filters) TSRMLS_CC);
+    }
+
+    if(APCG(lazy_functions)) {
+        APCG(lazy_function_table) = emalloc(sizeof(HashTable));
+        zend_hash_init(APCG(lazy_function_table), 0, NULL, NULL, 0);
+    }
+    if(APCG(lazy_classes)) {
+        APCG(lazy_class_table) = emalloc(sizeof(HashTable));
+        zend_hash_init(APCG(lazy_class_table), 0, NULL, NULL, 0);
+    }
 
 #ifdef APC_FILEHITS
     ALLOC_INIT_ZVAL(APCG(filehits));
@@ -736,6 +925,16 @@ int apc_request_init(TSRMLS_D)
 
 int apc_request_shutdown(TSRMLS_D)
 {
+
+    if(APCG(lazy_class_table)) {
+        zend_hash_destroy(APCG(lazy_class_table));
+        efree(APCG(lazy_class_table));
+    }
+    if(APCG(lazy_function_table)) {
+        zend_hash_destroy(APCG(lazy_function_table));
+        efree(APCG(lazy_function_table));
+    }
+
     apc_deactivate(TSRMLS_C);
 
 #ifdef APC_FILEHITS
